/*******************************************************************************
 *
 * ==============================================================================
 *
 * Copyright (c) 2001-2016 Primeton Technologies, Ltd. All rights reserved.
 * 
 * Created on 2017年4月11日 上午10:57:30
 *******************************************************************************/

package com.primeton.devops.common.service;

import java.util.Date;
import java.util.HashMap;
import java.util.Map;

import javax.annotation.PostConstruct;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.commons.lang.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import com.primeton.devops.common.DataContextThreadLocal;
import com.primeton.devops.common.util.CommonUtil;
import com.primeton.devops.common.util.ValidateUtil;
import com.primeton.devops.specs.api.pcm.service.IBusinessDictService;
import com.primeton.devops.specs.api.uc.service.IEmployeeService;
import com.primeton.devops.specs.api.uc.service.IUserService;
import com.primeton.devops.specs.constant.Constants;
import com.primeton.devops.specs.exception.DevOps;
import com.primeton.devops.specs.exception.DevOpsException;
import com.primeton.devops.specs.model.uc.Employee;
import com.primeton.devops.specs.model.uc.UserResult;
import com.primeton.ext.infra.security.BASE64Decoder;

import io.jsonwebtoken.Claims;
import io.jsonwebtoken.JwtParser;
import io.jsonwebtoken.Jwts;
import io.jsonwebtoken.SignatureAlgorithm;
import io.jsonwebtoken.impl.DefaultClaims;

/**
 * JwtToken工具类
 *
 * @author wuyuhou (mailto:wuyh@primeton.com)
 */
@Component
public class JwtTokenService {
	
	public static final String CLAIM_KEY_EMP_ID = "id";

	public static final String CLAIM_KEY_EMP_NAME = "name";

	public static final String CLAIM_KEY_TENENT_ID = "tenent";

	@Autowired
	private IEmployeeService employeeService;

	@Autowired
	private IBusinessDictService businessDictService;
	
	@Autowired
	private IUserService userService;

	private JwtParser jwtParser;
	
	private BASE64Decoder base64Decoder = new BASE64Decoder();
	
	@PostConstruct
	public void postConstruct() {
		jwtParser = Jwts.parser();
	}
	
	/**
	 * 生成token
	 * 
	 * @param empId 员工ID
	 * @param userName 用户名称
	 * @param tenantId 租户ID
	 * @return token
	 */
	public String generateToken(String empId, String userName, String empName, String tenantId) {
		ValidateUtil.assertNotNullOrEmpty(empId, "empId");
		ValidateUtil.assertNotNullOrEmpty(userName, "userName");
		ValidateUtil.assertNotNullOrEmpty(empName, "empName");
		if (StringUtils.isBlank(tenantId)) {
			tenantId = Constants.DEFAULT_TENANT_ID;
		}
		
		Claims claims = new DefaultClaims();
		claims.setSubject(userName);
		claims.put(CLAIM_KEY_EMP_ID, empId);
		claims.put(CLAIM_KEY_EMP_NAME, empName);
		claims.put(CLAIM_KEY_TENENT_ID, tenantId);
		return generateToken(claims);
	}
	
	/**
	 * 生成token
	 * 
	 * @param claims
	 * @return
	 */
	public String generateToken(Claims claims) {
		return generateToken(claims, getExpiration());
	}
	
	//失效时间获取
	private int getExpiration() {
		int expiration = 0;
		try {
			expiration = businessDictService.getConfigParamValue(null, "Pcm.TokenExpiration", int.class);
		} catch (Exception ignore) {}
		if (expiration <= 0) {
			expiration = 20;//默认20分钟
		}
		return expiration;
		
	}
	
	//获取secretKey
	private String getSecretKey() {
		try {
			return businessDictService.getConfigParamValue(null, "Pcm.TokenSecretKey", "devops-secret", String.class);
		} catch (Exception e) {
			return "devops-secret";
		}
	}
	
	/**
	 * 生成token
	 * 
	 * @param claims token项
	 * @param expiration 失效时间（分钟）
	 * @return
	 */
	public String generateToken(Claims claims, int expiration) {
		if (expiration <= 0) {
			return Jwts.builder().setClaims(claims).signWith(SignatureAlgorithm.HS512, getSecretKey()).compact();
		} else {
			return Jwts.builder().setClaims(claims).setExpiration(new Date(System.currentTimeMillis() + expiration * 1000 * 60)).signWith(SignatureAlgorithm.HS512, getSecretKey()).compact();
		}
	}
	
	/**
	 * 生成token
	 * 
	 * @param claims token项
	 * @param dueTime 到期时间
	 * @return
	 */
	public String generateToken(Claims claims, Date dueTime) {
		if (dueTime == null) {
			return Jwts.builder().setClaims(claims).signWith(SignatureAlgorithm.HS512, getSecretKey()).compact();
		} else {
			return Jwts.builder().setClaims(claims).setExpiration(dueTime).signWith(SignatureAlgorithm.HS512, getSecretKey()).compact();
		}
	}
	
	/**
	 * 解析token
	 * 
	 * @param token
	 * @return token信息
	 */
	public Claims parseToken(String token) {
		return jwtParser.setSigningKey(getSecretKey()).parseClaimsJws(token).getBody();
	}
	
	public void checkToken(HttpServletRequest request, HttpServletResponse response) throws Exception {
		String requestUri = request.getRequestURI();
		Map<String, String> queryMap = parseQuery(request.getQueryString());
		DataContextThreadLocal.setResponseCustomIncludeFields(CommonUtil.urlDecode((String)queryMap.get("responseCustomIncludeFields"), "UTF-8"));
		DataContextThreadLocal.setResponseCustomExcludeFields(CommonUtil.urlDecode((String)queryMap.get("responseCustomExcludeFields"), "UTF-8"));
		String projectId = request.getHeader(Constants.PROJECT_ID_HEAD_KEY);
		if (StringUtils.isBlank(projectId)) {
			projectId = queryMap.get(Constants.PROJECT_ID_HEAD_KEY);
		}
		if (StringUtils.isNotBlank(projectId)) {
			DataContextThreadLocal.setProjectId(projectId);
		}
		if (requestUri.contains("api/uc/users/login") 
				|| requestUri.contains("api/uc/users/tologin")
				|| requestUri.contains("api/uc/users/token/validate")
				|| requestUri.contains("user-doc")
				|| requestUri.contains("api/pcm/systeminfo")
				|| requestUri.contains("api/pcm/confg-parm-value")) {
			return;
		}
		String token = null;
		if (requestUri.contains("api/vcs/repos/actions/gitlabcallback")
				|| requestUri.contains("api/vcs/repos/actions/githubcallback")
				|| requestUri.contains("api/vcs/repos/actions/bitbucketcallback")) {
			token = generateToken(Constants.OUTER_EMP_ID, Constants.OUTER_USER_NAME, Constants.OUTER_EMP_NAME, null);
		} else {
			token = request.getHeader(Constants.AUTHORIZATION_HEAD_KEY);
			if (StringUtils.isBlank(token)) {
				token = queryMap.get(Constants.AUTHORIZATION_HEAD_KEY);
			}
		}
		if (StringUtils.isBlank(token) || "null".equalsIgnoreCase(token) || "undefined".equalsIgnoreCase(token)) {
			throw new DevOpsException(DevOps.Common.NOT_LOGIN, "need login.");
		}
		try {
			if (token.startsWith(Constants.BASIC_AUTHORIZATION_PREFIX)) {
				String userInfo = new String(base64Decoder.decodeBuffer(token.substring(Constants.BASIC_AUTHORIZATION_PREFIX.length())));
				int index = userInfo.indexOf(':');
				String userName = userInfo;
				String password = "";
				if (index != -1) {
					userName = userInfo.substring(0, index);
					password = userInfo.substring(index + 1);
				}
				DataContextThreadLocal.getUserInfo().setUserName(userName);
				UserResult userResult = userService.login(userName, password);
				response.setHeader(Constants.AUTHORIZATION_HEAD_KEY, userResult.getToken());
				return;
			}
			Claims claims = parseToken(token);
			DataContextThreadLocal.setToken(token);
			DataContextThreadLocal.setUserInfo((String)claims.get(JwtTokenService.CLAIM_KEY_EMP_ID), claims.getSubject(), (String)claims.get(JwtTokenService.CLAIM_KEY_EMP_NAME), (String)claims.get(JwtTokenService.CLAIM_KEY_TENENT_ID));
			if (claims.getExpiration() == null || requestUri.contains("api/uc/users/logout")) {
				return;
			}
			if (claims.getExpiration().before(new Date())) {
				throw new DevOpsException(DevOps.Common.NOT_LOGIN, "token is expiration, need relogin.");
			}
			if (!"false".equals(queryMap.get("refreshtoken"))) {
				response.setHeader(Constants.AUTHORIZATION_HEAD_KEY, generateToken(claims));
			}
		} catch (DevOpsException e) {
			throw e;
		} catch (Exception e) {
			throw new DevOpsException(e, DevOps.Common.NOT_LOGIN, "need relogin.");
		}
	}
	
	public void validateToken(String token) throws Exception {
		Claims claims = parseToken(token);
		if (claims.getExpiration() == null || claims.getExpiration().after(new Date())) {
			String userName = claims.getSubject();
			Employee employee = employeeService.getEmployee(userName);
			if (employee == null) {// 用户不存在
				throw new DevOpsException(DevOps.UC.ILLEGAL_TOKEN, token);
			}
		} else {
			throw new DevOpsException(DevOps.UC.ILLEGAL_TOKEN, token);
		}
	}
	
	private Map<String, String> parseQuery(String queryString) {
		Map<String, String> queryMap = new HashMap<String, String>();
		if (StringUtils.isBlank(queryString)) {
			return queryMap;
		}
		for (String query : queryString.split("&")) {
			String[] queryValue = query.split("=");
			if (queryValue.length == 2) {
				queryMap.put(queryValue[0], queryValue[1]);
			}
		}
		return queryMap;
	}
}
